import six
import keg_elements.encoding as encoding


def test_force_bytes():
    memory = memoryview('a') if six.PY2 else memoryview(b'a')

    assert encoding.force_bytes(memory) == b'a'
    assert encoding.force_bytes('a') == b'a'
    assert encoding.force_bytes(b'a') == b'a'

    try:
        encoding.force_bytes(True)
    except AttributeError:
        pass
    else:
        raise AssertionError('Expected force_bytes to throw')



import io
import os
import base64
from pathlib import Path

from keg_elements.encoding import force_bytes

import cryptography.fernet as fernet
import cryptography.hazmat.primitives.ciphers as ciphers
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import (
    constant_time,
    hashes,
    hmac,
    padding
)


def aes_cipher(key, iv, mode=None):
    algo = ciphers.algorithms.AES(key)
    mode = mode or ciphers.modes.CBC(iv)
    return ciphers.Cipher(algo, mode, backend=default_backend())


def aes_encryptor(key, mode=None):
    iv = os.urandom(ciphers.algorithms.AES.block_size // 8)
    cipher = aes_cipher(key, iv, mode=mode)
    return cipher.encryptor(), iv


def aes_decryptor(key, iv, mode=None):
    cipher = aes_cipher(key, iv=iv, mode=mode)
    return cipher.decryptor()


def fernet_cipher(key):
    return fernet.Fernet(base64.urlsafe_b64encode(key))


def encrypt(data, key):
    """
    Encrypts binary data using cryptography's default fernet algorithm
    :param data: plaintext data to encrypt
    :param key: encryption key
    :return: encrypted data as a fernet token (a signed, base64 encoded string)
    """
    return fernet_cipher(key).encrypt(data)


def decrypt(data, key):
    """
    Decrypts binary data using cryptography's default fernet algorithm
    :param data: a fernet token to decrypt
    :param key: encryption key
    :return: decrypted data
    """
    return fernet_cipher(key).decrypt(data)


def encrypt_str(data, key):
    """
    Encrypts a unicode string using cryptography's default fernet algorithm
    :param data: unicode string to encrypt
    :param key: encryption key
    :return: encrypted utf-8 string as a fernet token
    """
    bin_data = data.encode('utf-8')
    return encrypt(bin_data, key)


def decrypt_str(cipher_text, key):
    """
    Decrypts a unicode string using cryptography's default fernet algorithm
    :param cipher_text: a utf-8 string as a fernet token to decrypt
    :param key: encryption key
    :return: decrypted unicode string
    """
    bin_data = decrypt(cipher_text, key)
    return bin_data.decode('utf-8')


def encrypt_file(key, in_fpath, out_fpath=None, chunksize=64 * 1024):
    """ Encrypts a file using AES (CBC mode) with the
        given key.

        key:
            The encryption key - a string that must be
            either 16, 24 or 32 bytes long. Longer keys
            are more secure.

        in_fpath:
            Full path of the input file

        out_fpath:
            Full path of the output file.

            If None, '<in_fpath>.enc' will be used.

        chunksize:
            Sets the size of the chunk which the function
            uses to read and encrypt the file. Larger chunk
            sizes can be faster for some files and machines.
            chunksize must be divisible by 16.
    """
    if not out_fpath:
        out_fpath = in_fpath + '.enc'

    with open(in_fpath, 'rb') as infile, open(out_fpath, 'wb') as outfile:
        for chunk in encrypt_fileobj(key, infile, chunksize):
            outfile.write(chunk)

    return out_fpath


def encrypt_fileobj(key, in_fileobj, chunksize=64 * 1024):
    encryptor, iv = aes_encryptor(key)
    padder = padding.PKCS7(encryptor._ctx._cipher.block_size).padder()

    yield iv
    for chunk in iter(lambda: in_fileobj.read(chunksize), b''):
        data = padder.update(chunk)
        data = encryptor.update(data)
        yield data
    data = padder.finalize()
    yield encryptor.update(data)
    yield encryptor.finalize()


def decrypt_file(key, in_fpath, out_fpath=None, chunksize=24 * 1024):
    """ Decrypts a file using AES (CBC mode) with the
        given key. Parameters are similar to encrypt_file,
        with one difference: out_filename, if not supplied
        will be in_filename without its last extension
        (i.e. if in_filename is 'aaa.zip.enc' then
        out_filename will be 'aaa.zip')
    """
    if out_fpath is None:
        if Path(in_fpath).suffix == '.enc':
            out_fpath = Path(in_fpath).stem
        else:
            raise ValueError('If input file name doesn\'t end in ".enc" then output filename must '
                             'be given.')
    with open(in_fpath, 'rb') as infile, open(out_fpath, 'wb') as outfile:
        decrypt_fileobj(key, infile, outfile, chunksize)

    return out_fpath


def decrypt_bytesio(key, in_fpath, chunksize=24 * 1024):
    with open(in_fpath, 'rb') as infile:
        bytes_fobj = io.BytesIO()
        decrypt_fileobj(key, infile, bytes_fobj, chunksize)
    # prep for reading from the start of the file
    bytes_fobj.seek(0)
    return bytes_fobj


def decrypt_fileobj(key, in_fileobj, out_fileobj, chunksize):
    iv = in_fileobj.read(16)
    decryptor = aes_decryptor(key, iv)
    unpadder = padding.PKCS7(decryptor._ctx._cipher.block_size).unpadder()

    for chunk in iter(lambda: in_fileobj.read(chunksize), b''):
        data = decryptor.update(chunk)
        data = unpadder.update(data)
        out_fileobj.write(data)

    out_fileobj.write(unpadder.update(decryptor.finalize()))
    out_fileobj.write(unpadder.finalize())


def constant_time_compare(a, b):
    return constant_time.bytes_eq(a, b)


def salted_hmac(salt, value, secret):
    salt = force_bytes(salt)
    secret = force_bytes(secret)
    value = force_bytes(value)

    hasher = hashes.Hash(hashes.SHA512(), backend=default_backend())
    hasher.update(salt + secret)
    key = hasher.finalize()

    hmacer = hmac.HMAC(key, hashes.SHA512(), backend=default_backend())
    hmacer.update(value)
    return hmacer.finalize()



"""Common SQLAlchemy column types."""
import sqlalchemy as sa

from keg_elements import crypto


class TimeZoneType(sa.Unicode):
    """A column type for time zones, stored as a unicode string."""

    def __init__(self, length=None, **kwargs):  # Override any given length.
        super(TimeZoneType, self).__init__(length=255, **kwargs)


class EncryptedUnicode(sa.TypeDecorator):
    impl = sa.UnicodeText

    def __init__(self, *args, **kwargs):
        """
        Constructor for encrypted unicode type
        :param key: A bytes object containing the encryption key or a callable that returns the key
        :param encrypt: A callable that takes a unicode string and the encryption key as arguments
            and returns the encrypted data as a bytes object.
        :param decrypt: A callable that takes a bytes object and the encryption key as arguments
            and returns the decrypted data as a unicode string.
        """
        self._key = kwargs.pop('key')
        self._encrypt = kwargs.pop('encrypt', crypto.encrypt_str)
        self._decrypt = kwargs.pop('decrypt', crypto.decrypt_str)
        super(EncryptedUnicode, self).__init__(*args, **kwargs)

    @property
    def key(self):
        key_val = self._key() if callable(self._key) else self._key
        if len(key_val) < 32:
            raise ValueError('Key must be at least 32 bytes long')
        return key_val[:32]

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return self._encrypt(value, self.key).decode()

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self._decrypt(value.encode(), self.key)



from keg.db import db

import kegel_app.model.entities as ents
import pytz


class TestColumns(object):

    def test_testing_create_for_timezone_columns(self):
        obj = ents.ColumnTester.testing_create()
        assert obj.time_zone in pytz.common_timezones

        obj.time_zone = 'other'
        assert obj.time_zone == 'other'
        assert obj.__mapper__.columns.time_zone.type.length == 255

    def test_encrypted_null(self):
        ents.ColumnTester.delete_cascaded()
        obj = ents.ColumnTester.testing_create(encrypted1=None, encrypted2=None, encrypted3=None)
        assert obj.encrypted1 is None
        assert obj.encrypted2 is None
        assert obj.encrypted3 is None
        results = db.session.execute('select * from column_tester').fetchall()
        assert results[0].encrypted1 is None
        assert results[0].encrypted2 is None
        assert results[0].encrypted3 is None

    def test_store_encrypted(self):
        ents.ColumnTester.delete_cascaded()
        obj = ents.ColumnTester.testing_create(encrypted1='Foo', encrypted2='Bar', encrypted3='Baz')
        assert obj.encrypted1 == 'Foo'
        assert obj.encrypted2 == 'Bar'
        assert obj.encrypted3 == 'Baz'
        results = db.session.execute('select * from column_tester').fetchall()
        assert 'Foo' not in results[0].encrypted1
        assert 'Bar' not in results[0].encrypted2
        assert results[0].encrypted3 == 'Onm'

    def test_load_encrypted(self):
        ents.ColumnTester.delete_cascaded()
        db.session.execute('''
        insert into column_tester (encrypted1, encrypted2, encrypted3) values
            ('gAAAAABYVFHtgRTQJyzkIF5bQ0TiLA3Dm0kfDF5pvSLrZUBrpZiq_f3g3-jUUj8vPmTw2Uf3SW3u73fJNO9ER46xpW9_41IdaQ==',
            'gAAAAABYVFHtbibu6P-oroQ7UYj7g5oiZPqsbeAmxNAIz3wiEjc236bP7c2Ubsf1-S-ANt2r2WNQytxIfwxoE9VyNQmtihCCwA==',
            'Onm'
            )
        ''')
        obj = ents.ColumnTester.query.one()
        assert obj.encrypted1 == 'Foo'
        assert obj.encrypted2 == 'Bar'
        assert obj.encrypted3 == 'Baz'



import keg_elements.testing as testing


def test_dont_care():
    assert testing.DontCare() == True  # noqa
    assert (testing.DontCare() == True) is True  # noqa
    assert (testing.DontCare() == False) is True  # noqa
    assert (testing.DontCare() == object) is True  # noqa
    assert (testing.DontCare() != object) is False  # noqa
    assert (testing.DontCare()() == object) is True  # noqa
    assert (testing.DontCare()['thing'] == object) is True  # noqa
    assert (testing.DontCare()[1] == object) is True  # noqa



import codecs

import sqlalchemy as sa

from keg.db import db
import keg_elements.db.mixins as mixins
import keg_elements.db.columns as columns


many_things_mapper = db.Table(
    'many_things_mapper',
    sa.Column('thing_id', sa.Integer, sa.ForeignKey('things.id')),
    sa.Column('many_id', sa.Integer, sa.ForeignKey('many_things.id'))
)


class Thing(db.Model, mixins.DefaultMixin):
    __tablename__ = 'things'

    name = db.Column(db.Unicode(50), nullable=False)
    color = db.Column(db.Unicode)
    scale_check = db.Column(db.Numeric(8, 4))

    @sa.ext.hybrid.hybrid_property
    def name_and_color(self):
        return '{}-{}'.format(self.name, self.color)

    @name_and_color.expression
    def name_and_color(cls):
        return cls.name + sa.sql.literal('-') + cls.color


class RelatedThing(db.Model, mixins.DefaultMixin):
    __tablename__ = 'related_things'

    name = db.Column(db.Unicode(50), nullable=False)
    is_enabled = db.Column(db.Boolean, nullable=False, default=False,
                           server_default=sa.text('FALSE'))

    thing_id = sa.Column(sa.Integer, sa.ForeignKey(Thing.id), nullable=False)
    thing = sa.orm.relationship(lambda: Thing, backref='related_things')

    @classmethod
    def testing_create(cls, **kwargs):
        kwargs['thing'] = kwargs.get('thing') or Thing.testing_create()
        return super(RelatedThing, cls).testing_create(**kwargs)


class ManyToManyThing(db.Model, mixins.DefaultMixin):
    __tablename__ = 'many_things'

    name = db.Column(db.Unicode(50), nullable=False)

    things = sa.orm.relationship(lambda: Thing,
                                 secondary=many_things_mapper,
                                 backref='manys')


class ThingWithRequiredBoolean(db.Model, mixins.DefaultMixin):
    __tablename__ = 'required_boolean_table'

    nullable_boolean = db.Column(db.Boolean, nullable=True)
    required_boolean = db.Column(db.Boolean, nullable=False)
    required_boolean_with_default = db.Column(db.Boolean, nullable=False, default=False)
    required_boolean_with_server_default = db.Column(db.Boolean, nullable=False,
                                                     server_default='false')


class MultiplePrimaryKeys(db.Model, mixins.DefaultMixin):
    __tablename__ = 'multikey_table'

    other_pk = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Unicode(50))


def super_secure_encrypt(data, key):
    return codecs.encode(data, 'rot_13').encode()


def super_secure_decrypt(data, key):
    return codecs.encode(data.decode(), 'rot_13')


class ColumnTester(db.Model, mixins.DefaultMixin):
    __tablename__ = 'column_tester'

    time_zone = db.Column(columns.TimeZoneType(length=100))
    encrypted1 = db.Column(columns.EncryptedUnicode(key=lambda: b'a' * 32))
    encrypted2 = db.Column(columns.EncryptedUnicode(key=b'b' * 32))
    encrypted3 = db.Column(columns.EncryptedUnicode(
        key=b'c' * 32,
        encrypt=super_secure_encrypt,
        decrypt=super_secure_decrypt
    ))



from keg.app import Keg
from keg_elements.core import keg_element_blueprint


class KegElApp(Keg):
    import_name = 'kegel_app'
    keyring_enable = False
    use_blueprints = [keg_element_blueprint]
    db_enabled = True



from __future__ import absolute_import, unicode_literals

from pyquery import PyQuery as pq
from keg_elements.forms import FieldMeta, Form, ModelForm, SelectField
import keg_elements.forms as ke_forms
import six
from werkzeug.datastructures import MultiDict
import wtforms as wtf
from wtforms import validators
import wtforms_components

import kegel_app.model.entities as ents


class FormBase(object):
    form_cls = None
    entity_cls = None

    def compose_meta(self, fields_meta_cls=None, **data):
        assert self.entity_cls is not None, 'must set entity_cls on {}'\
            .format(self.__class__.__name__)

        class _ComposedForm(ModelForm):
            class Meta:
                model = self.entity_cls

            if fields_meta_cls:
                FieldsMeta = fields_meta_cls
            else:
                class FieldsMeta:
                    __default__ = FieldMeta
        return _ComposedForm(MultiDict(data))

    def create_form(self, **data):
        form_cls = data.pop('form_cls', self.form_cls)
        return form_cls(MultiDict(data))

    def assert_invalid(self, **data):
        form = self.create_form(**data)
        assert not form.validate()
        return form

    def assert_valid(self, **data):
        form = self.create_form(**data)
        assert form.validate(), form.errors
        return form


class FruitForm(Form):
    fruit = SelectField(validators=[validators.Optional()],
                        choices=[('apple', 'apple'), ('banana', 'banana')])
    letter = SelectField(validators=[validators.Optional()],
                         choices=[('a', 'a'), ('b', 'b'), ('', 'blank')], add_blank_choice=False)
    letter2 = SelectField(validators=[validators.InputRequired()], choices=[('a', 'a'), ('b', 'b')])
    number = SelectField(validators=[validators.Optional()], choices=[(1, '1'), (0, '0')])
    numstr = SelectField(validators=[validators.Optional()], choices=[('1', '1'), ('0', '0')],
                         coerce=six.text_type)


class TestSelectField(FormBase):
    form_cls = FruitForm

    def test_blank_choice_rendering(self):
        form = FruitForm(csrf_enabled=False)
        fruit_options = pq(form.fruit())('option')
        assert fruit_options.length == 3
        assert fruit_options.eq(0).attr('value') == ''
        assert fruit_options.eq(1).attr('value') == 'apple'

        letter_options = pq(form.letter())('option')
        assert letter_options.length == 3

    def test_blank_choice_submit(self):
        form = self.assert_valid(**{'fruit': '', 'letter': '', 'letter2': 'a'})

        assert form.fruit.data is None
        assert form.letter.data == ''

    def test_blank_choice_required(self):
        form = self.assert_invalid(**{'letter2': ''})
        assert form.letter2.errors == ['This field is required.']

        form = self.assert_valid(**{'letter2': 'b'})

    def test_blank_choice_coerce(self):
        self.assert_valid(**{'letter2': 'a', 'number': '1'})

        # make sure we can override the int helper by passing through an explicit coerce
        self.assert_valid(**{'letter2': 'a', 'numstr': '1'})


class TestRequiredBoolRadioField(FormBase):
    class RequiredBoolMockForm(Form):
        is_competent = ke_forms.RequiredBoolRadioField()
        statement_is = ke_forms.RequiredBoolRadioField(true_label='True', false_label='False')

    form_cls = RequiredBoolMockForm

    def test_required_bool_radio_rendering(self):
        form = self.RequiredBoolMockForm(csrf_enabled=False)
        is_competent_labels = pq(form.is_competent())('label')
        assert is_competent_labels.length == 2
        assert is_competent_labels.eq(0).text() == 'Yes'
        assert is_competent_labels.eq(1).text() == 'No'

        statement_is_labels = pq(form.statement_is())('label')
        assert statement_is_labels.length == 2
        assert statement_is_labels.eq(0).text() == 'True'
        assert statement_is_labels.eq(1).text() == 'False'

    def test_required_bool_radio_is_required(self):
        form = self.assert_invalid()
        assert form.is_competent.errors == ['This field is required.']
        assert form.statement_is.errors == ['This field is required.']

        form = self.assert_invalid(**{'is_competent': 'True'})
        assert form.is_competent.errors == []
        assert form.statement_is.errors == ['This field is required.']

        form = self.assert_invalid(**{'statement_is': 'True'})
        assert form.is_competent.errors == ['This field is required.']
        assert form.statement_is.errors == []

        form = self.assert_valid(**{'is_competent': 'False', 'statement_is': 'False'})


class TestDefaultTypeOfRequiredBooleanField(FormBase):
    entity_cls = ents.ThingWithRequiredBoolean

    def test_field_types(self):
        form = self.compose_meta(csrf_enabled=False)
        assert type(form.nullable_boolean) == wtf.fields.BooleanField
        assert type(form.required_boolean) == ke_forms.RequiredBoolRadioField
        assert type(form.required_boolean_with_default) == wtf.fields.BooleanField
        assert type(form.required_boolean_with_server_default) == ke_forms.RequiredBoolRadioField


class TestFieldMeta(FormBase):
    entity_cls = ents.Thing

    def test_required_unspecified(self):
        form = self.compose_meta()
        assert form.name.flags.required
        assert not form.color.flags.required

    def test_required_true(self):
        class Temp:
            __default__ = FieldMeta(required=True)
        form = self.compose_meta(Temp)
        assert form.name.flags.required
        assert form.color.flags.required

    def test_required_false(self):
        class Temp:
            __default__ = FieldMeta(required=False)
        form = self.compose_meta(Temp)
        assert not form.name.flags.required
        assert not form.color.flags.required

    def test_widget_no_override(self):
        form = self.compose_meta(csrf_enabled=False)
        assert type(form.color.widget) == wtforms_components.widgets.TextInput

    def test_widget_override(self):
        class WidgetOverrideFieldsMeta:
            __default__ = FieldMeta
            color = FieldMeta(widget=wtf.widgets.TextArea())

        form = self.compose_meta(fields_meta_cls=WidgetOverrideFieldsMeta, csrf_enabled=False)
        assert type(form.color.widget) == wtf.widgets.TextArea

    def test_extra_validators(self):
        class ExtraValidatorsFieldsMeta:
            def _is_roy(form, field):
                if field.data not in {'red', 'orange', 'yellow'}:
                    raise wtf.validators.ValidationError('Not a ROY color')

            __default__ = FieldMeta
            color = FieldMeta(extra_validators=[_is_roy])

        form = self.compose_meta(fields_meta_cls=ExtraValidatorsFieldsMeta, csrf_enabled=False,
                                 name='Test', color='red')
        assert form.validate()

        form = self.compose_meta(fields_meta_cls=ExtraValidatorsFieldsMeta, csrf_enabled=False,
                                 name='Test', color='muave')
        assert not form.validate()
        assert set(form.color.errors) == {'Not a ROY color'}


class TestValidators(FormBase):
    entity_cls = ents.Thing

    def test_numeric_scale_check(self):
        # too many places
        form = self.compose_meta(scale_check='0.11111')
        form.validate()
        assert form.scale_check.errors == ['Field must have no more than 4 decimal places.']

        # equal to data type
        form = self.compose_meta(scale_check='0.1111')
        form.validate()
        assert form.scale_check.errors == []

        # fewer places than data type
        form = self.compose_meta(scale_check='0.111')
        form.validate()
        assert form.scale_check.errors == []

        # integer
        form = self.compose_meta(scale_check='5')
        form.validate()
        assert form.scale_check.errors == []

        # test that other validation happens when type is wrong
        form = self.compose_meta(scale_check='aaa')
        form.validate()
        assert form.scale_check.errors == ['Not a valid decimal value']


class FeaturesForm(Form):
    name = wtf.StringField(validators=[validators.required()])
    color = wtf.StringField()


class NumbersSubForm(wtf.Form):
    number = wtf.StringField(validators=[validators.required()])
    color = wtf.StringField()


class NumbersForm(Form):
    numbers = wtf.FieldList(wtf.FormField(NumbersSubForm), min_entries=2)
    numbers2 = wtf.FieldList(wtf.StringField(), min_entries=2)


class TestFieldsToDict(FormBase):

    def test_simple(self):
        self.form_cls = FeaturesForm
        form = self.assert_invalid(color='blue')
        expected = {
            'csrf_token': {'data': '', 'errors': []},
            'name': {'data': '', 'errors': ['This field is required.']},
            'color': {'data': 'blue', 'errors': []},
        }
        assert form.fields_todict() == expected

    def test_simple_json(self):
        self.form_cls = FeaturesForm
        form = self.assert_invalid(color='blue')
        expected = {
            'csrf_token': {'data': '', 'errors': []},
            'name': {'data': '', 'errors': ['This field is required.']},
            'color': {'data': 'blue', 'errors': []},
        }
        fields_data = form.fields_todict()
        assert fields_data == expected

    def test_field_to_dict_field(self):
        form = NumbersSubForm()
        form.validate()
        assert ke_forms.field_to_dict(form.number) == \
            {'data': None, 'errors': ['This field is required.']}

    def test_field_to_dict_select(self):
        form = FruitForm(letter='b')
        form.validate()
        assert ke_forms.field_to_dict(form.letter) == \
            {'data': 'b', 'errors': []}
        assert ke_forms.field_to_dict(form.letter2) == \
            {'data': None, 'errors': ['This field is required.']}

    def test_field_to_dict_fieldlist(self):
        form = NumbersForm()
        form.validate()
        expected = [{'data': None, 'errors': []}, {'data': None, 'errors': []}]
        assert ke_forms.field_to_dict(form.numbers2) == expected

    def test_field_to_dict_form_fieldlist(self):
        data = {
            'numbers-0-number': '123',
            'numbers-1-color': 'blue',
        }
        form = NumbersForm(MultiDict(data))
        form.validate()
        expected = [
            {'number': {'data': '123', 'errors': []}, 'color': {'data': '', 'errors': []}},
            {'number': {'data': '', 'errors': ['This field is required.']},
                'color': {'data': 'blue', 'errors': []}},
        ]
        assert ke_forms.field_to_dict(form.numbers) == expected

    def test_nested(self):
        self.form_cls = NumbersForm
        data = {
            'numbers-0-number': '123',
            'numbers-1-color': 'blue'
        }
        form = self.assert_invalid(**data)
        expected = {
            'csrf_token': {'data': '', 'errors': []},
            'numbers': [
                {'number': {'data': '123', 'errors': []}, 'color': {'data': '', 'errors': []}},
                {'number': {'data': '', 'errors': ['This field is required.']},
                    'color': {'data': 'blue', 'errors': []}},
            ],
            'numbers2': [{'data': '', 'errors': []}, {'data': '', 'errors': []}]
        }
        fields_dict = form.fields_todict()
        assert fields_dict == expected, fields_dict


class TestFormLevelValidation(FormBase):
    class MyForm(Form):
        num1 = wtf.IntegerField()
        num2 = wtf.IntegerField()
        num3 = wtf.IntegerField()

        @ke_forms.form_validator
        def equal_42(self):
            if self.num1.data + self.num2.data != 42:
                raise wtf.ValidationError('Does not add up')

        @ke_forms.form_validator
        def in_order(self):
            if not (self.num1.data <= self.num2.data <= self.num3.data):
                raise wtf.ValidationError('Out of order')

    form_cls = MyForm

    def test_form_valid(self):
        form = self.assert_valid(num1=5, num2=37, num3=100)
        assert form.form_errors == []
        assert form.all_errors == ({}, [])

    def test_form_invalid(self):
        form = self.assert_invalid(num1=40, num2=3, num3=50)
        assert form.form_errors == ['Does not add up', 'Out of order']
        assert form.all_errors == ({}, ['Does not add up', 'Out of order'])

    def test_stop_validation_with_error(self):
        class StopValidationForm(Form):
            s1 = wtf.StringField()
            s2 = wtf.StringField()

            @ke_forms.form_validator
            def fields_equal(self):
                if self.s1.data != self.s2.data:
                    raise wtf.validators.StopValidation('not equal')

            @ke_forms.form_validator
            def other_validator(self):
                assert False, 'Validation should have stopped'  # pragma: no cover

        form = self.assert_invalid(form_cls=StopValidationForm, s1='v1', s2='v2')
        assert form.form_errors == ['not equal']
        assert form.all_errors == ({}, ['not equal'])

    def test_stop_validation_no_error(self):
        class StopValidationForm(Form):
            s1 = wtf.StringField()
            s2 = wtf.StringField()

            @ke_forms.form_validator
            def fields_equal(self):
                if self.s1.data != self.s2.data:
                    raise wtf.validators.StopValidation

            @ke_forms.form_validator
            def other_validator(self):
                assert False, 'Validation should have stopped'  # pragma: no cover

        form = self.assert_valid(form_cls=StopValidationForm, s1='v1', s2='v2')
        assert form.form_errors == []
        assert form.all_errors == ({}, [])

    def test_invalid_with_field_errors(self):
        class InvalidFieldsForm(Form):
            s1 = wtf.StringField(validators=[wtf.validators.Length(max=3)])
            s2 = wtf.StringField()

            @ke_forms.form_validator
            def fields_equal(self):
                if self.s1.data != self.s2.data:
                    raise wtf.ValidationError('not equal')

        form = self.assert_invalid(form_cls=InvalidFieldsForm, s1='1234', s2='4321')
        assert form.form_errors == ['not equal']
        assert form.all_errors == (
            {'s1': ['Field cannot be longer than 3 characters.']}, ['not equal'])

    def test_do_not_validate_with_field_errors(self):
        class InvalidFieldsForm(Form):
            s1 = wtf.StringField(validators=[wtf.validators.Length(max=3)])
            s2 = wtf.StringField()

            @ke_forms.form_validator(only_when_fields_valid=True)
            def my_validator(self):
                assert False, 'Validation should not run'  # pragma: no cover

        form = self.assert_invalid(form_cls=InvalidFieldsForm, s1='1234', s2='4321')
        assert form.form_errors == []
        assert form.all_errors == (
            {'s1': ['Field cannot be longer than 3 characters.']}, [])

    def test_validators_inherited(self):
        class SubclassForm(self.MyForm):
            @ke_forms.form_validator
            def num3_is_even(self):
                if self.num3.data % 2:
                    raise wtf.ValidationError('Num3 is odd')

            @ke_forms.form_validator
            def equal_42(self):
                if self.num1.data * self.num2.data != 42:
                    raise wtf.ValidationError('Does not compute')

        form = self.assert_invalid(num1=7, num2=5, num3=51, form_cls=SubclassForm)
        assert form.form_errors == ['Out of order', 'Num3 is odd', 'Does not compute']
        assert form.all_errors == ({}, ['Out of order', 'Num3 is odd', 'Does not compute'])

        form = self.assert_valid(num1=6, num2=7, num3=50, form_cls=SubclassForm)
        assert form.form_errors == []
        assert form.all_errors == ({}, [])



from io import BytesIO
from pathlib import Path

import cryptography
import pytest
from blazeutils import randchars
from cryptography.hazmat.primitives import ciphers

from keg_elements import crypto


CRYPTO_KEY = b'abcdefghijklmnopqrstuvwxyz123456'


def test_encrypt():
    data = b'\x01\x02\x03\x04\x05'
    encrypted = crypto.encrypt(data, CRYPTO_KEY)
    assert data not in encrypted
    assert encrypted != crypto.encrypt(data, CRYPTO_KEY)


def test_decrypt():
    data = b'\x01\x02\x03\x04\x05'
    assert crypto.decrypt(
        b'gAAAAABYVEWYMFW8Nyg91RlWubjVe_B82Q7AIyNDemSQSylUnJefKZ3k9smRdEP9-rD2vyNirrXqWCZ3wN1gk-Fu'
        b'gcdkwGuelQ==',
        CRYPTO_KEY
    ) == data
    assert crypto.decrypt(
        b'gAAAAABYVEXjvHp-r0FnZYJ5t3_p1H9TDMU6lGOAAeHerYkHR40XHCfEJl1FEI0w6WsmwW5Gc1976LVyf46jX4K3'
        b'BPfWkY8aNA==',
        CRYPTO_KEY
    ) == data

    with pytest.raises(cryptography.fernet.InvalidToken):
        crypto.decrypt(
            b'gAAAAABYVEXjvHp-r0FnZYJ5t3_p1H9TDMU6lGOAAeHerYkHR40XHCfEJl1FEI0w6WsmwW5Gc1976LVyf46j'
            b'X4K3BPfWkY8aNA==',
            b'a' * 32
        )


def test_encrypt_str():
    s = randchars()
    encrypted = crypto.encrypt_str(s, CRYPTO_KEY)
    assert s.encode() not in encrypted
    assert encrypted != crypto.encrypt_str(s, CRYPTO_KEY)


def test_decrypt_str():
    assert crypto.decrypt_str(
        b'gAAAAABYVEJVKXq6WlWn0wo3DDakIr6nwtHtk6llws4uJxgVvoBPLfhNRZc5eiWJEEAsEvtUjzQBqfaNscmqy2U'
        b'h0J21n1TyDA==',
        CRYPTO_KEY
    ) == 'foo'
    assert crypto.decrypt_str(
        b'gAAAAABYVENwMqAohDnzfTAZy6P4SV1Z8fOkREMNd52Oey1SeVE5aZcgcJBomEVFORJjMvh5LIUEQ2vgeVI-bi'
        b'CDjboq37uOaw==',
        CRYPTO_KEY
    ) == 'foo'


def test_constant_time_compare():
    # Known cases
    assert crypto.constant_time_compare(b'a', b'a')
    assert not crypto.constant_time_compare(b'aa', b'c')
    try:
        assert crypto.constant_time_compare('a', True)
    except TypeError:
        pass
    else:
        raise AssertionError('Should have raised a TypeError')

    # Fuzz Test
    rando1 = randchars(200).encode()
    rando2 = randchars(200).encode()
    rando3 = randchars(201).encode()
    assert crypto.constant_time_compare(rando1, rando1)
    assert crypto.constant_time_compare(rando2, rando2)
    assert crypto.constant_time_compare(rando3, rando3)
    assert not crypto.constant_time_compare(rando1, rando2)
    assert not crypto.constant_time_compare(rando1, rando3)


class TestCryptors:
    def enc_algo(self, encryptor):
        return encryptor._ctx._cipher.name, encryptor._ctx._mode.name

    def test_aes_encryptor_defaults(self):
        enc1, iv1 = crypto.aes_encryptor(CRYPTO_KEY)
        enc2, iv2 = crypto.aes_encryptor(CRYPTO_KEY)
        assert self.enc_algo(enc1) == ('AES', 'CBC')
        assert self.enc_algo(enc2) == ('AES', 'CBC')

        assert iv1 != iv2
        assert len(iv1) == 16
        assert len(iv2) == 16

    def test_aes_encryptor_mode(self):
        enc, iv = crypto.aes_encryptor(CRYPTO_KEY, mode=ciphers.modes.ECB())
        assert self.enc_algo(enc) == ('AES', 'ECB')

    def test_aes_decryptor_defaults(self):
        dec1 = crypto.aes_decryptor(CRYPTO_KEY, b'a' * 16)
        dec2 = crypto.aes_decryptor(CRYPTO_KEY, b'a' * 16)
        assert self.enc_algo(dec1) == ('AES', 'CBC')
        assert self.enc_algo(dec2) == ('AES', 'CBC')

    def test_aes_decryptor_mode(self):
        dec = crypto.aes_decryptor(CRYPTO_KEY, b'a' * 16, mode=ciphers.modes.ECB())
        assert self.enc_algo(dec) == ('AES', 'ECB')

    def test_encrypt_decrypt(self):
        data = b'a' * 32
        enc, iv = crypto.aes_encryptor(CRYPTO_KEY)
        enc_data = enc.update(data) + enc.finalize()

        # check CBC is functioning
        assert enc_data[:16] != enc_data[16:]

        dec = crypto.aes_decryptor(CRYPTO_KEY, iv)
        dec_data = dec.update(enc_data) + dec.finalize()
        assert dec_data == data


class TestStreamEncryption:
    @pytest.mark.parametrize("bytes_count", range(0, 100))
    def test_encrypt_decrypt_range(self, tmpdir, bytes_count):
        """
            The point of the test here is to do the encrypt/decrypt dance for a range of byte
            lengths that cross the 16 byte boundary and ensure we don't have corner cases that
            don't work.
        """
        value = b'b' * bytes_count
        fobj = tmpdir.join('{}-bytes.txt'.format(bytes_count))
        fobj.write(value)
        enc_fpath = crypto.encrypt_file(CRYPTO_KEY, fobj.strpath)

        bio_fobj = crypto.decrypt_bytesio(CRYPTO_KEY, enc_fpath)
        assert bio_fobj.read() == value

    @pytest.mark.parametrize("chunk_size", range(16, 300, 16))
    def test_encrypt_chunk_size(self, tmpdir, chunk_size):
        value = b'b' * 23
        fobj = tmpdir.join('{}-chunk.txt'.format(chunk_size))
        fobj.write(value)
        enc_fpath = crypto.encrypt_file(CRYPTO_KEY, fobj.strpath, chunksize=chunk_size)

        bio_fobj = crypto.decrypt_bytesio(CRYPTO_KEY, enc_fpath)
        assert bio_fobj.read() == value

    @pytest.mark.parametrize("chunk_size", range(1, 100, 1))
    def test_decrypt_chunk_size(self, tmpdir, chunk_size):
        value = b'b' * 23
        fobj = tmpdir.join('{}-chunk.txt'.format(chunk_size))
        fobj.write(value)
        enc_fpath = crypto.encrypt_file(CRYPTO_KEY, fobj.strpath)

        bio_fobj = crypto.decrypt_bytesio(CRYPTO_KEY, enc_fpath, chunk_size)
        assert bio_fobj.read() == value

    def test_encrypt_explicit_out_fpath(self, tmpdir):
        fobj = tmpdir.join('16-bytes.txt')
        fobj.write('b' * 16)
        enc_fpath = crypto.encrypt_file(CRYPTO_KEY, fobj.strpath, tmpdir.join('out.enc').strpath)

        bio_fobj = crypto.decrypt_bytesio(CRYPTO_KEY, enc_fpath)
        assert bio_fobj.read() == b'b' * 16

    def test_decrypt_bytesio(self, tmpdir):
        fobj = tmpdir.join('tempfile.txt')
        fobj.write('just some text')
        enc_fpath = crypto.encrypt_file(CRYPTO_KEY, fobj.strpath)

        bio_fobj = crypto.decrypt_bytesio(CRYPTO_KEY, enc_fpath)
        assert bio_fobj.read() == b'just some text'

    def test_decrypt_file(self, tmpdir):
        fobj = tmpdir.join('tempfile.txt')
        fobj.write('just some text')
        enc_fpath = crypto.encrypt_file(CRYPTO_KEY, fobj.strpath)

        # implicit name for output file
        out_fpath = crypto.decrypt_file(CRYPTO_KEY, enc_fpath)
        assert Path(out_fpath).open().read() == 'just some text'

        # explicit name for output file
        fobj = tmpdir.join('tempfile-encrypted.txt')
        out_fpath = crypto.decrypt_file(CRYPTO_KEY, enc_fpath, fobj.strpath)
        assert Path(out_fpath).open().read() == 'just some text'

        # Input file name doesn't in ".enc" but no output file name given.
        with pytest.raises(ValueError) as excinfo:
            fobj = tmpdir.join('wrong-name.txt')
            crypto.decrypt_file(CRYPTO_KEY, fobj.strpath)
        assert 'doesn\'t end in ".enc" then' in str(excinfo.value)

    def test_decrypt_without_truncate(self, tmpdir):
        fobj = tmpdir.join('tempfile.txt')
        fobj.write('just some text, and more for arguments sake')
        enc_fpath = crypto.encrypt_file(CRYPTO_KEY, fobj.strpath)

        # sys.stdout isn't really stdout, it's something py.test uses for capturing stdout, so
        # create our own container for the output:
        bio = BytesIO()

        def notruncate(*args, **kwargs):
            assert False, 'Truncate should not be called by decrypt_fileobj'  # pragma: no cover

        bio.truncate = notruncate

        with open(enc_fpath, 'rb') as infile:
            crypto.decrypt_fileobj(CRYPTO_KEY, infile, bio, 24 * 1024)



def base36_to_int(s):
    """
    Convert a base 36 string to an int. Raise ValueError if the input won't fit
    into an int.

    https://git.io/vydSi
    """
    # To prevent overconsumption of server resources, reject any
    # base36 string that is longer than 13 base36 digits (13 digits
    # is sufficient to base36-encode any 64-bit integer)
    if len(s) > 13:
        raise ValueError("Base36 input too large")
    return int(s, 36)


def int_to_base36(i):
    """Convert an integer to a base36 string.

    https://git.io/vydS1
    """
    char_set = '0123456789abcdefghijklmnopqrstuvwxyz'
    if i < 0:
        raise ValueError("Negative base36 conversion input.")
    if i < 36:
        return char_set[i]
    b36 = ''
    while i != 0:
        i, n = divmod(i, 36)
        b36 = char_set[n] + b36
    return b36



import keg_elements.decorators as decor


class TestKeywordOptional:
    class optional_execute:
        def __init__(self):
            self.called = False

        def __call__(self):
            self.called = True

    def test_keyword_optional_runs_with_true_pass(self):
        run_me = self.optional_execute()

        @decor.keyword_optional('test', before=run_me)
        def wrapped():
            pass

        wrapped(test=True)
        assert run_me.called is True

    def test_keyword_optional_doesnt_run_with_false_pass(self):
        run_me = self.optional_execute()

        @decor.keyword_optional('test', before=run_me)
        def wrapped():
            pass

        wrapped(test=False)
        assert run_me.called is False

    def test_keyword_optional_passes_keyword_when_told(self):
        run_me = self.optional_execute()

        @decor.keyword_optional('test', before=run_me, keep_keyword=True)
        def wrapped(test):
            return test

        assert wrapped(test=True) is True
        assert run_me.called is True

    def test_keyword_runs_after(self):
        run_me = self.optional_execute()

        @decor.keyword_optional('test', after=run_me)
        def wrapped():
            pass

        wrapped(test=True)
        assert run_me.called is True

    def test_when_missing_executes(self):
        run_me = self.optional_execute()

        @decor.keyword_optional('test', after=run_me, when_missing=True)
        def wrapped():
            pass

        wrapped(test=True)
        assert run_me.called is True



import six


def force_bytes(s, encoding='utf-8', errors='strict'):
    """Force objects to byte respresentation"""
    # Handle the common case first for performance reasons.
    if isinstance(s, bytes):
        if encoding == 'utf-8':
            return s
        else:
            return s.decode('utf-8', errors).encode(encoding, errors)
    elif isinstance(s, memoryview):
        return s.tobytes() if six.PY2 else bytes(s)
    else:
        return s.encode(encoding, errors)



from decimal import Decimal

import jinja2
from wtforms import ValidationError


def numeric(form, field):
    try:
        int(field.data)
    except ValueError:
        raise ValidationError('Value is not numeric.')


class NumberScale(object):
    def __init__(self, scale=-1, message=None):
        self.scale = scale
        if not message:
            message = u'Field must have no more than {} decimal places.'.format(scale)
        self.message = message

    def __call__(self, form, field):
        # if the data is not decimal, don't do anything with it (could be an alpha string)
        if not isinstance(field.data, Decimal):
            return
        # use decimal's quantization to see if it's equal to field data at the full scale.
        #   If it isn't, user entered too many decimal places
        if field.data != field.data.quantize(Decimal('0.{}1'.format('0' * (self.scale - 1)))):
            raise ValidationError(self.message)


class ValidateUnique(object):
    """Ensure the field's value is unique from other values

    This validator requires a ``get_object_by_field(field)`` on the form which
    returns an object (if one exists) or None to check for uniqueness.

    Pass ``object_html_link`` to this validator's ``__init__`` to get an error
    message which contains a hyperlink to the existing object.

    .. note ::

        Combine with ``wtforms.validators.optional`` when a field is nullable to
        allow for None values.
    """
    field_flags = ('unique',)

    def __init__(self, object_html_link=None):
        self.object_html_link = object_html_link

    def __call__(self, form, field):
        obj = getattr(form, 'obj')
        other = form.get_object_by_field(field)

        both_exist = None not in (obj, other)
        same_record = both_exist and other == obj
        another_exists_with_value = obj is None and other  # new obj with existing object with value

        if (both_exist and not same_record) or another_exists_with_value:
            link = (' to {}.'.format(self.object_html_link(other))
                    if hasattr(self, 'object_html_link') else '.')
            msg = jinja2.Markup('This value must be unique but is already assigned'
                                '{}'.format(link))
            raise ValidationError(msg)
        return True
